#ifndef BqW_autoptr_h
#define BqW_autoptr_h

/* This is Bisqwit's generic autoptr template file.
 * The same file is used in many different projects.
 *
 * autoptr.hh version 1.3.3
 */

template <typename T>
class autoptr
{
    inline void Forget() { if(p) { p->_ptr_lost_you(); if(p->_ptr_is_dead()) delete p; } }
    inline void Have(T *const a) const { if(a) a->_ptr_got_you(); }
    inline void Set(T *const a) { Have(a); Forget(); p = a; }
    inline void Birth() { Have(p); }
    T *p;
public:
    autoptr() : p(0) { }
    autoptr(T *const a) : p(a) { Birth(); }
    autoptr(const autoptr &a) : p(&*a) { Birth(); }
    
    // To enable boolean tests with this class: if(tmp) {...}
    inline operator const bool() const { return p; }
    inline const bool operator! () const { return !p; }
    
    // Act like if you were a pointer
    inline operator T *() const { return p; }
    
    // Dereferencing
    inline T &operator* () const { return *p; }
    inline T *operator-> () const { return p; }
    
    // Assigning
    autoptr &operator= (T *const a) { Set(a); return *this; }
    autoptr &operator= (const autoptr &a) { Set(&*a); return *this; }

    ~autoptr() { Forget(); }
};

template <typename T>
class autoeqptr : public autoptr<T>
{
public:
    autoeqptr() { }
    autoeqptr(T *const a) : autoptr<T> (a) { }
    template<typename K> inline bool operator== (K s) const { return **this == *s; }
    template<typename K> inline bool operator< (K s) const { return **this < *s; }
};

// autoptr must only use ptrable classes.
class ptrable
{
    mutable unsigned long _ptr_ref_num;
public:
    inline void _ptr_got_you() const { ++_ptr_ref_num; }
    inline void _ptr_lost_you() const { --_ptr_ref_num; }
    // _ptr_lost_you may not do "delete this",
    // because the destructor is not virtual.
    inline const bool _ptr_is_dead() const { return !_ptr_ref_num; }
public:
    ptrable() : _ptr_ref_num(0) { }
    ptrable(const ptrable &) : _ptr_ref_num(0) { }
};

#endif
